from common import featurecountsdf
import sys

df = featurecountsdf(sys.argv[1])

mapped = {s.split(":")[0] : int(s.split(":")[1]) for s in sys.argv[2].split(",")} # comma separated list of bamfname:mappedreads

#TODO: operate on all of the row's count cells at once as an array

for c in df.columns:
    if c in mapped:
        df[c+".RPKM"] = df[c]
        for row,reads in df[c+".RPKM"].iteritems():
            length = df.ix[row]["Length"]
            df.set_value(row,c+".RPKM", (10**9 * reads) / (length * mapped[c]))

df.to_csv("outas",sep="\t")



from Bio import SeqIO
from Bio.Seq import Seq
import pandas as pd
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
import glob,os,math,sys,json,pysam,operator
from gtf_parsing import load_gtf_as_dataframe

#pd.options.display.mpl_style = 'default'

#TODO: add strand to the gbk and gtf processors (for plotting the subfeatures
#TODO add ability to plot subfeatures as bars under the x axis

def parsearguments():
    epilog = """============================== Alea jacta est =============================="""
    parser = argparse.ArgumentParser(epilog=epilog,formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('ALNFILE',help="bam/sam format alignment file")
    parser.add_argument('FEATURESFILE',help="File from which to extract the features")
    parser.add_argument('OUTDIR',help="Output directory")
    parser.add_argument('--outprefix',help="Prefix for the output file names",default="")
    parser.add_argument('--features',help="Comma separated list of features to extract")
    parser.add_argument('--normfeat',help="Name of the subfeature to normalise all reads to")
    parser.add_argument('--normnreads',help="Normalise to the total number of aligned reads (per library)",action="store_true")
    parser.add_argument('--normval',help="Normalisation factor",default=None,type=float) #TODO improve behaviour
    parser.add_argument('--stranded', help="Generate a stranded plot", action="store_true")
    parser.add_argument('--binsize', help="Bin size to use for the histogram", type=int,default=10)
    parser.add_argument('--offsetup',help="Portion of sequence to show after the feature. Accepts a number (e.g. 1000) of nts, or a length multiplier (e.g. 1.5x)",default="0")
    parser.add_argument('--offsetdown',help="Portion of sequence to show before the feature. Accepts a number (e.g. 1000) of nts, or a length multiplier (e.g. 1.5x)",default="0")
    parser.add_argument('--log', help="Generate the y axis in log scale", action="store_true")
    parser.add_argument('--mmapweight', help="Weight the reads by the number of times it maps.", action="store_true")
    parser.add_argument('--nmaps', help="Maximum number of multi maps a read can have and still be counted", type=int, default=1)
    parser.add_argument('--length', help="Restrict counted reads to this length", type=int, default=None)
    parser.add_argument('--5pnt', help="Restrict counted reads to the ones with this 5p nt", default=None)

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    return(vars(parser.parse_args()))

def processbed(fname,featnames=None):
    features = []

    with open(fname, "rU") as f:
        for r in f:
            row = r.split()
            chrname = row[0]
            start = int(row[1])
            end = int(row[2])
            name = row[3]
            strand = row[5]
            if not featnames or name in featnames:
                features.append({"strand":strand,"name":name,"chr":chrname,"start":start,"end":end,"subfeatures":[]})

    return(features)

def processgbk(fname,featnames=None):
    features = []

    records = SeqIO.parse(open(fname, "rU"), "gb")
    for record in records:
        if not featnames or (record.name in featnames or record.id in featnames):
            subfeatures = [] 
            for subfeature in record.features:
                f = {}
                f["start"] = int(subfeature.location.start)
                f["end"] = int(subfeature.location.end)
                f["label"] = subfeature.qualifiers["label"][0]
                if "color" in subfeature.qualifiers:
                    f["color"] = subfeature.qualifiers["color"][0]
                if "fwcolor" in subfeature.qualifiers:
                    f["fwcolor"] = subfeature.qualifiers["fwcolor"][0]
                if "rvcolor" in subfeature.qualifiers:
                    f["rvcolor"] = subfeature.qualifiers["rvcolor"][0]
                subfeatures.append(f)
            features.append({"name":record.name,"start":1,"end":len(record.seq),"subfeatures":subfeatures})

    return(features)

def processgtf(gtffname,feattypes=['gene','exon'],featnames=None):
    #ensembl compatible...
    genes = []
    gene = None
    df = load_gtf_as_dataframe(gtffname)
    for row in df[df["feature"].isin(feattypes)].iterrows():
        data = row[1]
        if "gene_name" not in data:
            data["gene_name"] = data["gene_id"]
        ftype = data["feature"]
        if ftype == "gene" and (not featnames or (data["gene_name"] in featnames or data["gene_id"] in featnames)):
            if gene:
                genes.append(gene)
            gene = {"name":data["gene_name"],"id":data["gene_id"],"chr":data["seqname"],"start":data["start"],"end":data["end"],"subfeatures":[]}
        elif ftype == "exon":
            if gene:
                gene["subfeatures"].append({"label":"exon" + data["exon_number"],"start":data["start"],"end":data["end"]})

    genes.append(gene) #last gene would miss the filter otherwise
                
    return(genes)

def checkfilters(r,length,nt5p,nmaps):
    length = False if length and len(r.seq) != length else True
    if r.is_reverse:
        seq = Seq(r.seq)
        seq = seq.reverse_complement()
    else:
        seq = r.seq
    nt = False if nt5p and seq[0] != nt5p else True
    try:
        mult = True if dict(r.get_tags())["NH"] <= nmaps else False
    except KeyError:
        #print("WARNING: NH tag not present. Ignoring multi-mapper restriction")
        mult = True
    return mult and length and nt

def fhist(fname,feature,offsetup=0,offsetdown=0,normfeat=None,normnreads=False,normval=None,binsize=1,stranded=False,length=None,nt5p=None,nmaps=1,mmapweight=False):
    read_weights_fw = read_weights_rv = None

    bamfile = pysam.AlignmentFile(fname)

    reflen = feature["end"] - feature["start"]

    nbins = int(reflen/binsize)

    if offsetup and "x" in offsetup:
        offsetup = int(reflen * float(offsetup.replace("x","")))
    else:
        offsetup = int(offsetup)

    if offsetdown and "x" in offsetdown:
        offsetdown = int(reflen * float(offsetdown.replace("x","")))
    else:
        offsetdown = int(offsetdown)

    reflen += offsetdown + offsetup

    try:
        refname = feature["name"]
    except KeyError:
        refname = feature["id"]

    try:
        featname = feature["chr"]
    except KeyError:
        featname = feature["name"]

    if normfeat:
        for sf in feature["subfeatures"]:
            if sf["label"] == normfeat:
                normfeat = sf

    start = feature["start"]-offsetdown
    end = feature["end"]+offsetup

    if stranded:
        read_starts_fw = [read.reference_start + offsetdown - feature["start"] + 2 for read in bamfile.fetch(featname,start,end,multiple_iterators=True) if checkfilters(read,length=length,nmaps=nmaps,nt5p=nt5p) and not read.is_reverse]
        read_starts_rv = [read.reference_start + offsetdown - feature["start"] + 2 for read in bamfile.fetch(featname,start,end,multiple_iterators=True) if checkfilters(read,length=length,nmaps=nmaps,nt5p=nt5p) and read.is_reverse]

        if mmapweight:
            try:
                read_weights_fw = [1/dict(read.get_tags())["NH"] for read in bamfile.fetch(featname,start,end,multiple_iterators=True) if checkfilters(read,length=length,nmaps=nmaps,nt5p=nt5p) and not read.is_reverse]
                read_weights_rv = [1/dict(read.get_tags())["NH"] for read in bamfile.fetch(featname,start,end,multiple_iterators=True) if checkfilters(read,length=length,nmaps=nmaps,nt5p=nt5p) and read.is_reverse]
            except KeyError: # no multimapper info in the bamfile
                read_weights_fw = [1 for read in bamfile.fetch(featname,start,end,multiple_iterators=True) if checkfilters(read,length=length,nmaps=nmaps,nt5p=nt5p) and not read.is_reverse]
                read_weights_rv = [1 for read in bamfile.fetch(featname,start,end,multiple_iterators=True) if checkfilters(read,length=length,nmaps=nmaps,nt5p=nt5p) and read.is_reverse]
    else:
        read_starts_fw = [read.reference_start + offsetdown - feature["start"] + 2 for read in bamfile.fetch(featname,start,end,multiple_iterators=True) if checkfilters(read,length=length,nmaps=nmaps,nt5p=nt5p)]
        read_starts_rv = [0 for read in bamfile.fetch(featname,start,end,multiple_iterators=True)]

        if mmapweight:
            try:
                read_weights_fw = [1/dict(read.get_tags())["NH"] for read in bamfile.fetch(featname,start,end,multiple_iterators=True) if checkfilters(read,length=length,nmaps=nmaps,nt5p=nt5p)]
            except KeyError:
                read_weights_fw = [1 for read in bamfile.fetch(featname,start,end,multiple_iterators=True) if checkfilters(read,length=length,nmaps=nmaps,nt5p=nt5p)]

            read_weights_rv = [1 for read in bamfile.fetch(featname,start,end,multiple_iterators=True)]

    counts_fw,bins_fw = np.histogram(np.array(read_starts_fw),bins=nbins,range=(0,reflen),weights=read_weights_fw)
    counts_rv,bins_rv = np.histogram(np.array(read_starts_rv),bins=nbins,range=(0,reflen),weights=read_weights_rv)

    if stranded:
        counts_rv *= -1

    if not stranded or sum(counts_fw) > sum(counts_rv): #which bins to use when calculating width, etc
        refbins = bins_fw
    else:
        refbins = bins_rv

    #TODO: improve this logic. They are not mutually exclusive at the argument level
    if normval:
        counts_fw = counts_fw / normval
        counts_rv = counts_rv / normval
    elif normnreads:
        try:
            nalnreads = sum(1/dict(read.get_tags())["NH"] for read in bamfile.fetch(featname,start,end,multiple_iterators=True))
        except KeyError:
            nalnreads = sum(1 for read in bamfile.fetch(featname,start,end,multiple_iterators=True))

        counts_fw = counts_fw / nalnreads
        counts_rv = counts_rv / nalnreads
    elif normfeat:
        tmp = sum(r >= normfeat["start"] and r <= normfeat["end"] for r in read_starts_fw)
        tmp += sum(r >= normfeat["start"] and r <= normfeat["end"] for r in read_starts_rv)
        counts_fw = counts_fw / tmp
        counts_rv = counts_rv / tmp

    width = (refbins[1] - refbins[0])
    center = (refbins[:-1] + refbins[1:]) / 2

    counts = {"fw":counts_fw}
    if counts_rv.any():
        counts["rv"] = counts_rv

    return counts,refbins,center,width,feature
        
def plot(counts,refbins,center,width,feature,fheight=6,fwidth=14,ylims=(-1,1),title=""):
    if "rv" not in counts:
        stranded = False
    else:
        stranded = True
    
    reflen = feature["end"] - feature["start"]
    
    counts_fw = counts["fw"]
    counts_rv = counts["rv"]
    
    miny,maxy = ylims
    
    if max(counts_fw) > ylims[1]:
        maxy = max(counts_fw)
    if stranded and min(counts_rv) < ylims[0]:
        miny = min(counts_rv)

    if stranded:
        fwcolordef = "blue"
    else:
        fwcolordef = "black"
    rvcolordef = "red"

    fig, ax = plt.subplots(figsize=(fwidth,fheight))
    fwax = ax.bar(center, counts_fw, align='center', width=width)
    if stranded:
        rvax = ax.bar(center, counts_rv, align='center', width=width)

    for (bin_start,patch) in zip(refbins,fwax.patches):
        patch.set_edgecolor("none")
        patch.set_color(fwcolordef)
        if "subfeatures" in feature:
            for sf in feature["subfeatures"]:
                try:
                    fwcolor = sf["fwcolor"]
                except KeyError:
                    try:
                        fwcolor = sf["color"]
                    except KeyError:
                        if stranded:
                            fwcolor = "blue"
                        else:
                            fwcolor = "black"

                if bin_start >= sf["start"] and bin_start < sf["end"]:
                    patch.set_color(fwcolor)
                    break

    if stranded:
        for (bin_start,patch) in zip(refbins,rvax.patches):
            patch.set_edgecolor("none")
            patch.set_color(rvcolordef)
            if "subfeatures" in feature:
                for sf in feature["subfeatures"]:
                    try:
                        rvcolor = sf["rvcolor"]
                    except KeyError:
                        try:
                            rvcolor = sf["color"]
                        except KeyError:
                            rvcolor = "red"

                    if bin_start >= sf["start"] and bin_start < sf["end"]:
                        patch.set_color(rvcolor)
                        break

    ax.set_title(title)
    ax.set_xlim(0,reflen)
    ax.set_ylim(miny,maxy)

    return(fig)

if __name__ == "__main__":
    import argparse,sys
    from common import *

    args = parsearguments()

    if args["features"]:
        featlist = args["features"].split(",")
    else:
        featlist = None

    featsfname = args["FEATURESFILE"]

    fext = os.path.splitext(featsfname)[1]

    if fext in [".gbk",".gb",".ape"]:
        features = processgbk(featsfname,featnames=featlist)
    elif fext == ".gtf":
        features = processgtf(featsfname,featnames=featlist)
    elif fext == ".bed":
        features = processbed(featsfname,featnames=featlist)
    else:
        raise ValueError("Could not identify features file type from extension {}".format(fext))


    for feature in features:

        counts,refbins,center,width,feature = fhist(args["ALNFILE"],feature,offsetup=args["offsetup"],offsetdown=args["offsetdown"],normfeat=args["normfeat"],stranded=args["stranded"],normnreads=args["normnreads"],normval=args["normval"],binsize=args["binsize"],length=args["length"],nt5p=args["5pnt"],nmaps=args["nmaps"],mmapweight=args["mmapweight"])

        fig = plot(counts,refbins,center,width,feature,title=args["ALNFILE"])

        fig.savefig(args["ALNFILE"].replace(".bam","_{}.svg".format(feature["name"])))



def parsearguments():
    epilog = """============================== Alea jacta est =============================="""
    parser = argparse.ArgumentParser(epilog=epilog,formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('INFNAME',help="Path to the SAM/BAM/fastq/fasta file containing the reads")
    parser.add_argument('OUTFNAME',help="Output file name.")
    parser.add_argument('--lengths',type=int,metavar='N',help="Sequence lengths to extract",nargs='*')
    parser.add_argument('--reqnts',type=str,metavar='XN',help="Required nucleotides at a certain positions. For a T in position 10, pass 10T",nargs="*")
    parser.add_argument('--nmaps', help="Maximum number of multi maps", type=int, default=1)
    parser.add_argument('--nmism', help="Maximum number of mismatches", type=int, default=0)
    parser.add_argument('--endtoend', help="Require end to end alignments (no clipping whatsoever)", action="store_true")

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    return(vars(parser.parse_args()))

def checkreqnts(record,reqnts):
    if reqnts:
        seq = Seq(str(record.seq))
        try:
            if record.is_reverse:
                seq = seq.reverse_complement()
        except AttributeError:
            pass #it's not a xam file
        for nt in reqnts:
            if str(seq)[nt[1]-1] != nt[0]:
                return False
    return True

def isxam(fext):
    return fext in [".bam",".sam"]

def filteralignments(alignments,lengths=None,reqnts=None,nmism=0,nmaps=1,endtoend=False):
    for aln in alignments:
        if (not lengths or aln.alen in lengths) and checkreqnts(aln,reqnts) and (not endtoend or len(aln.cigar) == 1):
            tags = dict(aln.get_tags())
            try:
                if tags["nM"] <= nmism and tags["NH"] <= nmaps:
                    yield aln
            except KeyError:
                print("WARNING: bam file doesn't seems to contain some tags")
                yield aln

def parsereqnts(reqnts):
    try:
        reqnts = [[nt[-1],int(nt[0:-1])] for nt in args["reqnts"]]
    except TypeError:
        reqnts = []

    return reqnts

if __name__ == "__main__":
    import argparse,sys,os
    from common import *
    from tstk.io import openfastx
    import gzip
    import pysam
    from Bio import SeqIO
    from Bio.Seq import Seq
    from Bio.SeqRecord import SeqRecord

    args = parsearguments()

    lengths = args["lengths"]
    reqnts = parsereqnts(args["reqnts"])
    reqpos = [i[1] for i in reqnts]

    if reqpos and lengths and max(reqpos) > max(lengths):
        print("WARNING: some of your fixed positions fall beyond the maximum requested length")

    ifpathnoext, ifext = os.path.splitext(args["INFNAME"])
    ofpathnoext, ofext = os.path.splitext(args["OUTFNAME"])

    if not isxam(ifext) and isxam(ofext):
        raise TypeError("ERROR: cannot output BAM from a FASTX input")

    if isxam(ifext):
        infile = pysam.AlignmentFile(args["INFNAME"])

        if isxam(ofext):
            outfile = pysam.AlignmentFile(args["OUTFNAME"], mode="wb", template=infile)
            for aln in filteralignments(infile,lengths,reqnts,args["nmism"],args["nmaps"],args["endtoend"]):
                outfile.write(aln)
        else:
            ofpathnoext,ofext,oftype,ofh = openfastx(args["OUTFNAME"],mode="wt")
            for aln in filteralignments(infile,lengths,reqnts,args["nmism"],args["nmaps"],args["endtoend"]):
                seq = Seq(aln.seq)
                seq.name = seq.id = seq.description = aln.query_name
                if aln.is_reverse:
                    seq = seq.reverse_complement()
                record = SeqRecord(seq,id=aln.query_name,description="")
                SeqIO.write(record,ofh,oftype)

            ofh.close()
    else:
        fpathnoext,fext,ftype,fh = openfastx(args["INFNAME"])
        infile = SeqIO.parse(fh,ftype)

        ofpathnoext,ofext,oftype,ofh = openfastx(args["OUTFNAME"],mode="wt")

        for record in SeqIO.parse(fh,ftype):
            if (not lengths or len(record.seq) in lengths) and checkreqnts(record,reqnts):
                SeqIO.write(record,ofh,ftype)
        fh.close()
        ofh.close()



#converts a fasta file to a simplified annotation format file or BED file. Each sequence is converted into a single entry that takes an entire "chromosome"
# SAF files are accepted by featureCounts. Useful when building custom "genomes".
from Bio import SeqIO
import sys,os

def parsearguments():
    epilog = """============================== Alea jacta est =============================="""
    parser = argparse.ArgumentParser(epilog=epilog,formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('FASTAFILE',help="Path to the fasta file containing the reads")
    parser.add_argument('OUTFNAME',help="Output file name. Should have a .bed or .saf extension.")

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    return(vars(parser.parse_args()))

if __name__ == "__main__":
    import argparse,sys,collections
    from Bio import SeqIO

    args = parsearguments()
    fname,ext = os.path.splitext(args["OUTFNAME"])

    if ext == ".saf":
        rowt = "{name}\t{chrname}\t0\t{end}\t+\n"
    elif ext == ".bed":
        rowt = "{chrname}\t0\t{end}\t{name}\t.\t+\n"
    else:
        raise ValueError("Output file extension must be either .bed or .saf")

    with open(args["OUTFNAME"],"w") as fo:
        for record in SeqIO.parse(args["FASTAFILE"],"fasta"):
            fo.write(rowt.format(chrname = record.id,end = len(record.seq)-1,name = record.id))



def openfastx(fpath,mode="rt"):
    import os

    gz = False
    fpathnoext, fext = os.path.splitext(fpath)
    if fext == '.gz':
        gz = True
        fpathnoext, fext = os.path.splitext(fpathnoext)

    if fext.lstrip(".") in ['fasta','fa']:
        ftype = 'fasta'
    elif fext.lstrip(".") in ['fastq','fq']:
        ftype = 'fastq'
    else:
        raise ValueError("File type [fasta/fastq] could not be determined from the filename '{}'.".format(fpath))

    if gz:
        import gzip
        fh = gzip.open(fpath,mode)
    else:
        fh = open(fpath,mode)

    return fpathnoext,fext,ftype,fh

def parsefastx(fp): # this is a generator function
    last = None # this is a buffer keeping the last unprocessed line
    while True: # mimic closure; is it a bad idea?
        if not last: # the first record or a record following a fastq
            for l in fp: # search for the start of the next record
                if l[0] in '>@': # fasta/q header line
                    last = l[:-1] # save this line
                    break
        if not last: break
        name, seqs, last = last[1:].partition(" ")[0], [], None
        for l in fp: # read the sequence
            if l[0] in '@+>':
                last = l[:-1]
                break
            seqs.append(l[:-1])
        if not last or last[0] != '+': # this is a fasta record
            yield name, ''.join(seqs), None # yield a fasta record
            if not last: break
        else: # this is a fastq record
            seq, leng, seqs = ''.join(seqs), 0, []
            for l in fp: # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(seq): # have read enough quality
                    last = None
                    yield name, seq, ''.join(seqs); # yield a fastq record
                    break
            if last: # reach EOF before reading enough quality
                yield name, seq, None # yield a fasta record instead
                break

def openxam(fpath,mode="r",template=None):
    print("DONT USE THIS ANYMORE, JUST CALL pysam.AlignmentFile directly. Quitting now.")
    import pysam,sys,os

    if mode not in ['r','rb','w','wb']:
        raise ValueError("Unrecognized open mode '{}'".format(mode))

    fpathnoext, fext = os.path.splitext(fpath)

    if fext == ".bam" and len(mode) == 1:
        mode += "b"

    return(pysam.AlignmentFile(fpath, mode, template=template))



from tstk.common import *
import pysam
import sys

def parsearguments():
    import argparse

    description = """===================== Uber getter of overlapping reads ====================="""
    epilog = """============================== Alea jacta est =============================="""
    parser = argparse.ArgumentParser(description=description, epilog=epilog,formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('seqfile',help="Path to the bam file containing the reads")
    parser.add_argument('outfname',help="Output bam file name")
    parser.add_argument('--lengths',type=int,default=[21,22,23,24],metavar='N',help="Length of the + strand read",nargs="*")
    parser.add_argument('--offset',type=int,default=2,metavar='N',help="Offset of the reverse read (positive value => offset towards the forward read 3')")
    parser.add_argument('--chrname', help="Count reads for the specified 'chromosome' only (BAM exclusive)")

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    return(vars(parser.parse_args()))

def checksoutplus(read,lengths):
    return not read.is_reverse and read.alen in lengths

def checksoutminus(readp,readm,offset):
    return readm.is_reverse and read.alen == readp.alen and readp.reference_start - readm.reference_start == offset

def getchunks(l, n):
    return [l[i::n] for i in range(n)]

def processreads(seqfile,refpositions,offset):
    nreads = len(refpositions)
    res = []

    fi = pysam.AlignmentFile(seqfile)

    for i,fwpos in enumerate(refpositions):
        if __name__ == "__main__":
            sys.stderr.write("Processing read {}/{}\n".format(i+1,nreads))

        chrname = fwpos[0]
        start = fwpos[1]
        end = fwpos[2]

        reads = fi.fetch(reference=fi.getrname(chrname),start=start,end=end)

        dps = {"fw":[],"rv":[]}

        for read in reads:
            if read.alen == (end-start+1):
                if read.is_reverse:
                    if read.reference_start == start - offset:
                        dps["rv"].append(read)
                else:
                    if read.reference_start == start:
                        dps["fw"].append(read)

        if len(dps["rv"]) > 0:                
            res += dps["fw"]
            res += dps["rv"]

    fi.close()

    return(res)

def getrefpositions(seqfile,chrname,lengths):
    infile = pysam.AlignmentFile(seqfile)
    refpos = {(r.reference_id,r.reference_start,r.reference_start+r.qend-r.qstart-1) for r in infile.fetch(chrname) if checksoutplus(r,lengths)}
    infile.close()

    return refpos

def getdicerproducts(seqfile, chrname, lengths, offset):
    refpos = getrefpositions(seqfile,chrname,lengths)
    return processreads(seqfile,refpos,offset)

def main(seqfile, outfname, chrname, lengths, offset):
    infile = pysam.AlignmentFile(seqfile)
    outfile = pysam.AlignmentFile(outfname, "wb", template=infile)
    infile.close()

    for read in getdicerproducts(seqfile,chrname,lengths,offset):
        outfile.write(read)

    outfile.close()

if __name__ == "__main__":
    args = parsearguments()
    main(**args)



from tstk.io import openfastx
import gzip,sys,os
from Bio import SeqIO
from Bio.Seq import Seq

def getfiltseqs(fname):
    fpathnoext, fext = os.path.splitext(fname)
    if fext in [".bam",".sam"]:
        filt_seqs = [(record.query_name,str(record.seq)) for record in openxam(fname)]
    else:
        fpathnoext,fext,ftype,fh = openfastx(fname)
        filt_seqs = [(record.id,str(record.seq)) for record in SeqIO.parse(fh,ftype)]

    return(set(filt_seqs))

def gettargetseqs(fname):
    fpathnoext, fext = os.path.splitext(fname)
    if fext in [".bam",".sam"]:
        return openxam(fname,"r")
    else:
        fpathnoext,fext,ftype,fh = openfastx(fname)
        return SeqIO.parse(fh,ftype)

    return(set(filt_seqs))

def parsearguments():
    import argparse
    epilog = """============================== Alea jacta est =============================="""
    parser = argparse.ArgumentParser(epilog=epilog,formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('TARGETS',help="Path to the SAM/BAM/fastq/fasta file containing target sequences")
    parser.add_argument('FILTERS',help="Path to the SAM/BAM/fastq/fasta file containing filtering reads")
    parser.add_argument('OUTFNAME',help="Output file name.")
    parser.add_argument('--3ponly',help="Don't get sequences that have a 5' overhang",action="store_true")
    parser.add_argument('--output_filters',help="Output a file containing the matched filters",action="store_true",default=True)
    parser.add_argument('--output_targets',help="Output a file containing the matched targets",action="store_true")

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    return(vars(parser.parse_args()))

def filter_boyer(filter_seq,target_seqs,target_seqs_flat,target_seqs_flat_divs,only3p):
    import bisect

    matched_target_seqs = []
    matched_target_ids = []
    i = target_seqs_flat.find(filter_seq)
    while i != -1:
        p = bisect.bisect_right(target_seqs_flat_divs, i) - 1
        matched_target_seq = target_seqs[p]
        if not only3p or matched_target_seq[1].find(filter_seq) == 0:
            matched_target_seqs.append(matched_target_seq)
        i = target_seqs_flat.find(filter_seq, target_seqs_flat_divs[p] + len(matched_target_seq[1]))

    return matched_target_seqs

def filterseqs(target_seqs,filter_seqs,only3p=False):
    import re

    matched_target_seqs = []
    matched_target_ids=[]
    matching_filter_seqs = []

    target_seqs_flat = "#"+"#".join([s[1] for s in target_seqs]) 
    target_seqs_flat_divs = [m.start()+1 for m in re.finditer('#', target_seqs_flat)]

    nfilts = len(filter_seqs)
    for i,fs in enumerate(filter_seqs):
        m = filter_boyer(fs[1],target_seqs,target_seqs_flat,target_seqs_flat_divs,only3p) 
        matched_target_seqs += m
        if len(m) > 0:
            matching_filter_seqs.append(fs)

    return(set(matched_target_seqs),matching_filter_seqs)

def filterseqs_from_files(ftargets,ffilters):

    target_seqs = []

    for record in gettargetseqs(ftargets):
        try:
            seq = Seq(record.seq)
            if record.is_reverse:
                seq = seq.reverse_complement()
            target_seqs.append((record.query_name,str(seq)))
        except TypeError:
            seq = str(record.seq)
            target_seqs.append((record.id,str(seq)))

    filter_seqs = getfiltseqs(ffilters)

    return filterseqs(target_seqs,filter_seqs)

if __name__ == "__main__":
    args = parsearguments()

    if not args["OUTFNAME"].endswith(".fasta"):
        args["OUTFNAME"] += ".fasta"

    matched_target_seqs,matching_filter_seqs = filterseqs_from_files(args["TARGETS"],args["FILTERS"])

    if args["output_targets"]:
        ofpathnoext,ofext,oftype,ofh = openfastx("targets_" + args["OUTFNAME"],mode="wt")
        for id,seq in matched_target_seqs:
           ofh.write(">{}\n{}\n".format(id,seq))
        ofh.close()

    if args["output_filters"]:
        ofpathnoext,ofext,oftype,ofh = openfastx("filters_" + args["OUTFNAME"],mode="wt")
        for id,seq in matching_filter_seqs:
           ofh.write(">{}\n{}\n".format(id,seq))
        ofh.close()



import sys,glob,os

def countdups(fi,fo):
    #This function takes a collapsed fasta file (from fastx_collapser) and calculates the fractions of Unique reads, duplicate reads and singletons.
    #It then prints out the info to a tab-delimited text file that can be opened in Excel.

    #Opens a fasta file with the database sequences as specified by user. Must be non-interleaved.
    with open(fi, 'r') as SEQS:
        #Opens an output text file as specified by user
        with open(fo, 'w') as HEADERS:
            with open(fi.replace(".txt","_dupcount.csv"), 'w') as OUT:
                duplicates=0
                dupreadcount=0
                singletons=0
                totalreadcount=0
                HEADERS.write('UniqueNumber'+'\t'+'Numseqscollapsed')
                for line in SEQS:
                    line=line.rstrip()
                    if line[0:1] == '>':
                        line=line.strip('>')
                        cols=line.split('-')
                        if int(cols[1]) > 1:
                            duplicates+=1
                            dupreadcount+=int(cols[1])
                            totalreadcount+=int(cols[1])
                            HEADERS.write('\n'+'\t'.join(cols))
                        elif int(cols[1]) == 1:	
                            totalreadcount+=int(cols[1])
                            singletons+=1

                percentsingletons=float(singletons)/float(totalreadcount)
                percentduplicates=float(dupreadcount)/float(totalreadcount)
                percentuniquedups=float(duplicates)/float(dupreadcount)
                OUT.write('n_uniq_w_dups'+'\t'+'n_collapsed_dups'+'\t'+ 'n_singletons' + '\t' + 'total_reads' + '\t' + 'pct_singletons' + '\t' + 'pct_dups' + '\t' + 'pct_uniq_w_dups')
                OUT.write('\n' + str(duplicates) + '\t' + str(dupreadcount) + '\t' + str(singletons) + '\t' + str(totalreadcount) + '\t' + str(percentsingletons) + '\t' + str(percentduplicates) + '\t' + str(percentuniquedups))

for f in glob.glob("fastq/out/*.txt"):
    print("*************")
    print(f)
    fo = f.replace(".txt","_headers.txt")
    countdups(f,fo)
    os.remove(fo)



compnt = {'A':'T','C':'G','G':'C','T':'A','N':'N'}

revcomp = lambda x: ''.join([compnt[B] for B in x][::-1])

def featurecountsdf(fname,allcolumns=False):
    import pandas as pd

    df = pd.read_table(fname,sep="\t",header=1) 
    if not allcolumns:
        cols = ["Geneid"] + [c for c in df.columns[df.columns.get_loc("Length"):]]
        df = df[cols]

    return df



import pysam
import subprocess
from contextlib import contextmanager, closing

def parsearguments():
    epilog = """============================== Alea jacta est =============================="""
    parser = argparse.ArgumentParser(epilog=epilog,formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('BAMFILE',help="BAM file containing the reads")
    parser.add_argument('OUTFILE',help="Output file (wig)")
    parser.add_argument('--chrom',help="Limit the processing to a specific chromosome",default="all")
    parser.add_argument('--strand',help="Limit the processing to a specific strand",choices=["fw","rv"])
    parser.add_argument('--rpm', help="Write counts as reads per million", action="store_true")
    parser.add_argument('--neg', help="Negate counts", action="store_true")

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    return(vars(parser.parse_args()))

@contextmanager
def indexed_bam(bam_file):
    if not os.path.exists(bam_file + ".bai"):
        pysam.index(bam_file)
    bam = pysam.AlignmentFile(bam_file)
    yield bam
    bam.close()

def main(bam_file, outfile, chrom='all', start=0, end=None, rpm=False,strand=None):
    if start > 0:
        start = int(start) - 1
    if end is not None:
        end = int(end)
    regions = [(chrom, start, end)]

    if strand:
        n,x = os.path.splitext(outfile)
        outfile = "{}.{}{}".format(n,strand,x)

    out_handle = open(outfile, "w")
    with closing(out_handle):
        chr_sizes = write_bam_track(bam_file, regions, out_handle, rpm, strand)
    convert_to_bigwig(outfile, chr_sizes)

def write_bam_track(bam_file, regions, out_handle, rpm, strand=None):
    out_handle.write("track {}\n".format(" ".join(["type=wiggle_0", "name={}{}".format(os.path.splitext(os.path.split(bam_file)[-1])[0],"_"+strand if strand else ""), "visibility=full", ])))
    with indexed_bam(bam_file) as work_bam:
        if strand:
            if strand == "fw":
                goodstrand = lambda r: not r.is_reverse
            elif strand == "rv":
                goodstrand = lambda r: r.is_reverse
            else:
                raise ValueError("Bad strand value {}".format(strand))
        else:
            goodstrand = lambda r: True

        total = sum(1.0 / dict(r.get_tags())["NH"] for r in work_bam.fetch() if not r.is_unmapped) if rpm else None
        sizes = list(zip(work_bam.references, work_bam.lengths))
        if len(regions) == 1 and regions[0][0] == "all":
            regions = [(name, 0, length) for name, length in sizes]
        for chrom, start, end in regions:
            if end is None and chrom in work_bam.references:
                end = work_bam.lengths[work_bam.references.index(chrom)]
            assert end is not None, "Could not find {} in header".format(chrom)
            out_handle.write("variableStep chrom={}\n".format(chrom))
            for col in work_bam.pileup(chrom, start, end):
                n = sum(1.0 / dict(r.alignment.get_tags())["NH"] for r in col.pileups if goodstrand(r.alignment)) 
                if rpm:
                    n = float(n) / total * 1e6
                if args["neg"]:
                    n *= -1
                out_handle.write("%s %.1f\n" % (col.pos+1, n))

    return sizes

def convert_to_bigwig(wig_file, chr_sizes):
    bw_file = "{}.bw".format(os.path.splitext(wig_file)[0])
    size_file = "{}.sizes.txt".format(os.path.splitext(wig_file)[0])
    with open(size_file, "w") as out_handle:
        for chrom, size in chr_sizes:
            out_handle.write("%s\t%s\n" % (chrom, size))
    try:
        import shlex
        s = subprocess.Popen(shlex.split("wigToBigWig {} {} {}".format(wig_file,size_file,bw_file)),stdin=subprocess.PIPE,stdout=subprocess.PIPE)
        s.communicate()
    finally:
        os.remove(size_file)
    return bw_file

if __name__ == "__main__":
    import argparse,sys,os

    args = parsearguments()

    main(args["BAMFILE"],args["OUTFILE"],chrom=args["chrom"],rpm=args["rpm"],strand=args["strand"])



def parsearguments():
    import argparse,sys

    epilog = """============================== Alea jacta est =============================="""
    parser = argparse.ArgumentParser(epilog=epilog,formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('OUTDIR',help="Output directory")

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    return(vars(parser.parse_args()))

if __name__ == "__main__":
    import urllib.request
    import pandas as pd
    from time import strftime

    args = parsearguments()

    types = [['tRNA',6238164],['rRNA',6238163]]

    for t in types:
        url = "http://www.informatics.jax.org/marker/report.txt?mcv={}".format(t[1])
        data = pd.read_table(urllib.request.urlopen(url))

        data.columns = ['chr', 'start', 'end', 'cM', 'strand', 'MGI ID', 'Feature Type', 'symbol', 'name', 'Unnamed: 9', 'Unnamed: 10']
        data['score'] = 0
        data['chr'] = 'chr' + data['chr']
        data = data[['chr','start','end','symbol','score','strand']].dropna()
        data.loc[data['chr']=='chrMT','chr'] = 'chrM'
        data[['start','end']] = data[['start','end']].astype(int)

        outfname = "{}/MGI_{}_{}.bed".format(args["OUTDIR"],t[0],strftime("%Y%m%d"))
        data.to_csv(outfname,sep="\t",header=False,index=False)




def parsearguments():
    import argparse,sys

    description = """=========================== Peterplot generator ============================"""
    epilog = """============================== Alea jacta est =============================="""
    parser = argparse.ArgumentParser(description=description, epilog=epilog,formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('SEQFILE',help="Path to the fastq/collapsed fasta/bam file containing the reads")
    parser.add_argument('OUTFNAME',help="Output file name")
    parser.add_argument('--minlength',type=int,default=18,metavar='N',help="Minimum length to report on the x axis")
    parser.add_argument('--maxlength',type=int,default=33,metavar='N',help="Maximum length to report on the x axis")
    parser.add_argument('--ylims',metavar='N,N',help="Limits for the y axis, comma separated (lower,upper)")
    parser.add_argument('--title',default="",help="Title for the plot")
    parser.add_argument('--uncollapse',action="store_true",help="Uncollapse the reads (requires fasta headers formatted as 'id-count'")
    parser.add_argument('--normfw',type=float,help="Normalisation factor for the forward strand",default=None)
    parser.add_argument('--normrv',type=float,help="Normalisation factor for the reverse strand",default=None)
    parser.add_argument('--normnreads',action="store_true",help="Normalise reads to library size (or mapped reads in case of BAM)")
    parser.add_argument('--stranded', help="Generate a stranded plot", action="store_true")
    parser.add_argument('--fwonly', help="Plot only forward strand", action="store_true")
    parser.add_argument('--rvonly', help="Plot only reverse strand", action="store_true")
    parser.add_argument('--noN', help="Don't display the counts for the 'N' nucleotide", action="store_true")
    parser.add_argument('--3p', help="Count the 3' nucleotides instead of the 5'", action="store_true")
    parser.add_argument('--chrnames', help="Count reads for the specified 'chromosome' only (BAM exclusive)", nargs="*")
    parser.add_argument('--nmaps', help="Maximum number of multi maps a read can have and still be counted (BAM exclusive)", type=int, default=1)

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    return(vars(parser.parse_args()))

def writeplot(OUTFNAME,fig):
    fig.savefig(OUTFNAME)

def plot(counts,minlength=18,maxlength=33,title=None,ylims=None,stranded=False,noN=True,fwonly=False,rvonly=False,colours=None,isrna=True,legend=True):
    import matplotlib.pyplot as plt
    import pandas as pd
    import numpy as np
    import copy

    if not title:
        title = ""

    counts = copy.deepcopy(counts)

    if fwonly or rvonly:
        s = "rv" if fwonly else "fw"
        for l in counts[s]:
            for n in counts[s][l]:
                counts[s][l][n] = 0

    lengths = np.arange(minlength,maxlength+1)
    if stranded:
        lengths_rv = np.arange(minlength,maxlength+1)

    df = pd.DataFrame(counts["fw"])
    df.fillna(0,inplace=True)
    if stranded:
        df_rv = pd.DataFrame(counts["rv"]) * -1
        df_rv.fillna(0,inplace=True)

    fig, ax = plt.subplots()
    ax.set_title(title)
    if colours:
        nucs = [["A",colours["A"]],["C",colours["C"]],["G",colours["G"]],["T",colours["T"]]]
    else:
        nucs = [["A","red"],["C","blue"],["T","green"],["G","purple"]]
    if not noN:
        nucs += ["N","cyan"]

    bottom = [0] * len(df.columns)
    bottom_rv = [0] * len(df.columns)
    idx = np.arange(len(df.columns))
    width = 0.8

    for n in nucs:
        if n[0] in df.index:
            label = n[0] if not isrna or n[0] != "T" else "U"
            ax.bar(idx, df.loc[n[0]], color=n[1], bottom=bottom, width=width,label=label)
            bottom += df.loc[n[0]]
            if stranded:
                ax.bar(idx, df_rv.loc[n[0]], color=n[1], bottom=bottom_rv, width=width)
                bottom_rv += df_rv.loc[n[0]]

    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.tick_params(axis='both', direction='out')
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    plt.xticks(idx+width/2.,lengths)
    plt.xlim([0,len(lengths)])

    if ylims:
        plt.ylim(ylims)
    if legend:
        plt.legend()

    return(fig)

def getcounts(SEQFILE,minlength=18,maxlength=33,uncollapse=False,normfw=None,normrv=None,normnreads=False,stranded=False,noN=False,c3p=False,chrnames=None,nmaps=1,t2u=False,rvonly=False,fwonly=False):

    from tstk.io import openfastx
    from Bio import SeqIO
    from Bio.Seq import Seq
    import numpy as np
    import sys,os,re

    fpath = SEQFILE
    ext = os.path.splitext(fpath)[1]

    lengths = np.arange(minlength,maxlength+1)
    if stranded:
        lengths_rv = np.arange(minlength,maxlength+1)
    counts = {}

    if noN:
        counts = {key: {"A":0,"C":0,"T":0,"G":0} for key in lengths}
        counts_rv = {key: {"A":0,"C":0,"T":0,"G":0} for key in lengths}
    else:
        counts = {key: {"A":0,"C":0,"T":0,"G":0,"N":0} for key in lengths}
        counts_rv = {key: {"A":0,"C":0,"T":0,"G":0,"N":0} for key in lengths}

    if c3p:
        nucpos = -1
    else:
        nucpos = 0

    nrecords = 0
    if ext == ".bam":
        import pysam
        bamfile = pysam.AlignmentFile(fpath)
        if chrnames:
            reads = []
            for chrname in chrnames:
                reads += bamfile.fetch(chrname)
        else:
            reads = bamfile.fetch()

        for read in reads:
            tags = dict(read.get_tags())
            if "NH" not in tags or tags["NH"] <= nmaps:
                if read.is_reverse:
                    seq = Seq(read.seq)
                    seq = seq.reverse_complement()
                else:
                    seq = read.seq

                firstnuc = seq[nucpos]
                if not noN or firstnuc != "N":
                    if len(seq) in lengths:
                        if stranded and read.is_reverse:
                                counts_rv[len(seq)][firstnuc] += 1
                        else:
                                counts[len(seq)][firstnuc] += 1

        if normnreads:
            if chrnames:
                nrecords = 0 
                for chrname in chrnames:
                    nrecords += sum(1 for r in pysam.AlignmentFile(fpath).fetch(chrname) if dict(r.get_tags())["HI"] == 1)
            else:
                nrecords = sum(1 for r in pysam.AlignmentFile(fpath) if dict(r.get_tags())["HI"] == 1)
    else:
        checkheader = True

        fpathnoext,fext,ftype,fh = openfastx(fpath)

        for record in SeqIO.parse(fh,ftype):
            if uncollapse:
                if checkheader:
                    reid = re.compile("^\d+-\d+$")
                    if not reid.match(record.id):
                        print("WARNING: the fasta header doesn't seem to match the 'id-count' format. Ignoring uncollapse request.")
                        uncollapse = False
                    checkheader = False

                if ftype == "fastq":
                    print("WARNING: file format is fastq, ignoring uncollapse request")
                    uncollapse = False
                    count = 1
                else:
                    count = int(record.id.split("-")[1])
            else:
                count = 1

            nrecords += count

            firstnuc = record.seq[nucpos]
            if not noN or firstnuc != "N":
                if len(record.seq) in lengths:
                    counts[len(record.seq)][firstnuc] += count

    normfactor_fw = None
    normfactor_rv = None

    if normnreads:
        normfactor_fw = nrecords
        normfactor_rv = nrecords

    if normfw:
        if normnreads:
            print("WARNING: normalisation to library size specified together with specific normalisation factor. Using the specific factor for FW reads")
        normfactor_fw = normfw

    if normrv:
        if normnreads:
            print("WARNING: normalisation to library size specified together with specific normalisation factor. Using the specific factor for RV reads")
        normfactor_rv = normrv

    if normfactor_fw: 
        for l in counts:
            for n in counts[l]:
                counts[l][n] = counts[l][n] / normfactor_fw

    if stranded and normfactor_rv:
        for l in counts_rv:
            for n in counts_rv[l]:
                counts_rv[l][n] = counts_rv[l][n] / normfactor_rv

    try:
        fh.close()
    except NameError:
        pass

    if t2u:
        for l in counts:
            counts[l]["U"] = counts[l].pop("T")
        for l in counts_rv:
            counts_rv[l]["U"] = counts_rv[l].pop("T")

    counts = {"fw":counts,"rv":counts_rv}

    return counts

if __name__ == "__main__":
    import os
    import matplotlib as mpl
    mpl.use('Agg')

    args = parsearguments()

    if args["uncollapse"]:
        args["OUTFNAME"] = args["OUTFNAME"].replace(".collapsed","")

    outformats = ['svg','png','jpg','eps','raw','ps','pdf','tif','pgf']

    outdir,outfname = os.path.split(args["OUTFNAME"])
    outfnamenoext,outfext = os.path.splitext(outfname)
    outformat = outfext[1:]
    if outformat not in outformats:
        raise ValueError("Output format not recognised form file extension '{}'. Please use one of {}".format(outformat,outformats))

    title = args["title"] if args["title"] else os.path.basename(args["OUTFNAME"])
    ylims = [float(lim) for lim in ylims.split(",")] if args["ylims"] else None

    counts = getcounts(args["SEQFILE"],minlength=args["minlength"],maxlength=args["maxlength"],uncollapse=args["uncollapse"],normfw=args["normfw"],normrv=args["normrv"],normnreads=args["normnreads"],stranded=args["stranded"],noN=args["noN"],c3p=args["3p"],chrnames=args["chrnames"],nmaps=args["nmaps"],fwonly=args["fwonly"],rvonly=args["rvonly"])
    f = plot(counts,args["minlength"],args["maxlength"],title=title,ylims=ylims,stranded=args["stranded"],fwonly=args["fwonly"],rvonly=args["rvonly"])
    writeplot(args["OUTFNAME"],f)



from setuptools import setup

version = '0.2.1'

setup(
    name = 'tstk',
    version = version,
    packages = ['tstk'],
    description = 'A bunch of tools for sequencing data analysis',
    author = 'Tomás Di Domenico',
    author_email = 'tdido@tdido.com.ar',
    url = 'https://github.com/tdido/tstk',
    keywords = ['sequence','toolkit','bioinformatics'],
    install_requires=[
        'pandas',
        'pysam',
        'biopython',
        'matplotlib',
        'numpy'
    ],
    license = 'MIT',
    classifiers=[  # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ]
)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Parse an Ensembl GTF file into a dataframe, expanding all of its attributes
into their own columns,


Columns of a GTF file:

    seqname   - name of the chromosome or scaffold; chromosome names
                without a 'chr'
    source    - name of the program that generated this feature, or
                the data source (database or project name)
    feature   - feature type name. Current allowed features are
                {gene, transcript, exon, CDS, Selenocysteine, start_codon,
                stop_codon and UTR}
    start     - start position of the feature, with sequence numbering
                starting at 1.
    end       - end position of the feature, with sequence numbering
                starting at 1.
    score     - a floating point value indiciating the score of a feature
    strand    - defined as + (forward) or - (reverse).
    frame     - one of '0', '1' or '2'. Frame indicates the number of base pairs
                before you encounter a full codon. '0' indicates the feature
                begins with a whole codon. '1' indicates there is an extra
                base (the 3rd base of the prior codon) at the start of this feature.
                '2' indicates there are two extra bases (2nd and 3rd base of the
                prior exon) before the first codon. All values are given with
                relation to the 5' end.
    attribute - a semicolon-separated list of tag-value pairs (separated by a space),
                providing additional information about each feature. A key can be
                repeated multiple times.

(from ftp://ftp.ensembl.org/pub/release-75/gtf/homo_sapiens/README)
"""

from __future__ import print_function, division, absolute_import
import logging
from os.path import exists

import numpy as np
import pandas as pd


GTF_COLS = [
    'seqname',
    # Different versions of GTF use this column as
    # of: (1) gene biotype (2) transcript biotype or (3) the annotation source
    #
    # See: https://www.biostars.org/p/120306/#120321
    'second_column',
    'feature',
    'start',
    'end',
    'score',
    'strand',
    'frame',
    'attribute'
]

def pandas_series_is_biotype(series):
    """
    Hackishly infer whether a Pandas series is either from
    gene_biotype or transcript_biotype annotations by checking
    whether the 'protein_coding' biotype annotation is among its values.
    """
    return 'protein_coding' in series.values

def _read_gtf(filename):
    """
    Download GTF annotation data for given release (if not already present),
    parse it into a DataFrame, leave the attributes in a single string column
    """
    assert exists(filename)
    assert filename.endswith(".gtf") or filename.endswith(".gtf.gz"), \
        "Must be a GTF file (%s)" % filename
    compression = "gzip" if filename.endswith(".gz") else None
    df = pd.read_csv(
        filename,
        comment='#',
        sep='\t',
        names=GTF_COLS,
        na_filter=False,
        compression=compression)

    df['seqname'] = df['seqname'].map(str)

    # very old GTF files use the second column to store the gene biotype
    # others use it to store the transcript biotype and
    # anything beyond release 77+ will use it to store
    # the source of the annotation (e.g. "havana")
    if pandas_series_is_biotype(df['second_column']):
        # patch this later to either 'transcript_biotype' or 'gene_biotype'
        # depending on what other annotations are present
        column_name = 'biotype'
    else:
        column_name = 'source'
    df.rename(columns={'second_column' : column_name}, inplace=True)

    return df

def _attribute_dictionaries(df):
    """
    Given a DataFrame with attributes in semi-colon
    separated string, split them apart into per-entry dictionaries
    """
    # this crazy triply-nested comprehensions
    # is unfortunately the best way to parse
    # ~2 million attribute strings while still using
    # vaguely Pythonic code
    attr_dicts = [
        {
            key : value.replace("\"", "").replace(";", "")
            for (key,value) in (
                pair_string.split(" ", 2)
                for pair_string in attr_string.split("; ")
            )
        }
        for attr_string in df.attribute
    ]
    return attr_dicts

def _extend_with_attributes(df):
    n = len(df)
    extra_columns = {}
    column_order = []
    for i, attr_string in enumerate(df.attribute):
        # TODO: tokenize attribute strings properly
        # for now, catch mistaken semicolons by replacing "xyz;" with "xyz"
        # Required to do this since the GTF for Ensembl 78 has
        # gene_name = "PRAMEF6;"
        # transcript_name = "PRAMEF6;-201"
        attr_string = attr_string.replace(';\"', '\"').replace(";-", "-")

        pairs = (
            kv.strip().split(" ", 2)
            for kv in attr_string.split(";")
            # simplest entry must be at least three characters:
            # (key name, space, value)
            if len(kv) > 2
        )
        for k,v in pairs:
            if k not in extra_columns:
                extra_columns[k] = [""] * n
                column_order.append(k)
            extra_columns[k][i] = v

    # make a copy of the DataFrame without attribute strings.
    df = df.drop("attribute", axis=1)

    logging.info("Adding attribute columns: %s", column_order)
    for k in column_order:
        assert k not in df, "Column '%s' appears in GTF twice" % k
        df[k] = extra_columns[k]
        # delete from dictionary since these objects are big
        # and we might be runniung low on memory
        del extra_columns[k]
        # remove quotes around values
        df[k] = df[k].str.replace("\"", "")
    return df

# In addition to the required 8 columns and names and IDs genes & transcripts,
# there might also be annotations like 'transcript_biotype' but these aren't
# available for all species/releases.

REQUIRED_ATTRIBUTE_COLUMNS = [
    'gene_name',
    'gene_id',
    'gene_biotype',
    'transcript_name',
    'transcript_id',
]

def _dataframe_from_groups(groups, feature, extra_column_names=[]):
    """
    Helper function used to construct a missing feature such as 'transcript'
    or 'gene'. For example, a sufficiently old release might only have
    'exon' entries, but these were tagged with which transcript_id and gene_id
    they're associated with. Grouping by transcript_id lets you reconstruct
    the feature='transcript' entries which are normally present in later
    releases.
    """
    start = groups.start.min()
    end = groups.end.max()
    strand = groups.strand.first()
    seqname = groups.seqname.first()
    gene_name = groups.gene_name.first()
    gene_biotype = groups.gene_biotype.first()
    def pick_protein_id(candidates):
        for c in candidates:
            if c is not None and len(c) > 0:
                return c
        return None
    protein_id = groups.protein_id.apply(pick_protein_id)
    columns = [seqname, gene_biotype, start, end, strand, gene_name, protein_id]
    for column_name in extra_column_names:
        column = groups[column_name].first()
        columns.append(column)
    df = pd.concat(columns, axis=1).reset_index()

    # score seems to be always this value, not sure why it's in the GTFs
    df['score'] = '.'

    # frame values only make sense for CDS entries, but need this column
    # so we concatenate these rows with the rest of the Ensembl entries
    df['frame'] = '.'

    df['feature'] = feature
    return df

def reconstruct_gene_rows(df):
    gene_id_groups = df.groupby(['gene_id'])
    genes_df = _dataframe_from_groups(gene_id_groups, feature='gene')
    return pd.concat([df, genes_df], ignore_index=True)

def reconstruct_transcript_rows(df):
    transcript_id_groups = df.groupby(['transcript_id'])
    transcripts_df = _dataframe_from_groups(
        transcript_id_groups,
        feature='transcript',
        extra_column_names=['transcript_name']
    )
    return pd.concat([df, transcripts_df], ignore_index=True)

def reconstruct_exon_id_column(df, inplace=True):
    """
    Construct missing exon_id column for older GTFs by concatenating
    transcript_id and exon_number

        e.g. ENST00000400674.exon2

    While not useful for joining against external data, these exon_ids are
    needed for methods like ensembl_release.exon_ids_at_location.

    Parameters
    ----------

    df : DataFrame
        Must have columns 'transcript_id' and 'exon_number'

    """

    assert 'exon_id' not in df
    assert 'transcript_id' in df
    assert 'exon_number' in df

    if not inplace:
        df = df.copy()

    # missing values in dataframes indicated by NaN
    df['exon_id'] = np.nan

    # assign exon IDs to any entry with an exon number
    # this includes features = {'exon', 'CDS', 'start_codon', 'stop_codon'}
    exon_id_mask = ~df.exon_number.isnull()

    transcript_ids = df['transcript_id'][exon_id_mask]

    # convert floating value to ints (to get rid of decimal) and then str
    exon_numbers = df['exon_number'][exon_id_mask].astype(int).astype(str)

    df['exon_id'][exon_id_mask] = transcript_ids + '.exon' + exon_numbers
    return df

def load_gtf_as_dataframe(filename):
    logging.info("Reading GTF %s into DataFrame", filename)
    df = _read_gtf(filename)
    logging.info(
        "Extracting attributes for %d entries in GTF DataFrame", len(df))
    df = _extend_with_attributes(df)

    # due to the annoying ambiguity of the second GTF column,
    # figure out if an older GTF's biotype is actually the gene_biotype
    # or transcript_biotype

    if 'biotype' in df.columns:
        assert 'transcript_biotype' not in df.columns, \
            "Inferred 2nd column as biotype but also found transcript_biotype"

        # Initially we could only figure out if the 2nd column was either
        # the source of the annotation or some kind of biotype (either
        # a gene_biotype or transcript_biotype).
        # Now we disambiguate between the two biotypes by checking if
        # gene_biotype is already present in another column. If it is,
        # the 2nd column is the transcript_biotype (otherwise, it's the
        # gene_biotype)
        if 'gene_biotype' in df.columns:
            rename_to = 'transcript_biotype'
        else:
            rename_to = 'gene_biotype'
        df.rename(columns={'biotype' : rename_to}, inplace=True)

    for column_name in REQUIRED_ATTRIBUTE_COLUMNS:
        assert column_name in df.columns, \
            "Missing required column '%s', available: %s" % (
                column_name,
                list(sorted(df.columns)))

    # older Ensembl releases only had features:
    #   - exon
    #   - CDS
    #   - start_codon
    #   - stop_codon
    #
    # Might have to manually reconstruct gene & transcript entries
    # by grouping the gene_id and transcript_id columns of existing features

    distinct_features = df.feature.unique()

    if 'gene' not in distinct_features:
        logging.info("Creating entries for feature='gene'")
        df = reconstruct_gene_rows(df)

    if 'transcript' not in distinct_features:
        logging.info("Creating entries for feature='transcript'")
        df = reconstruct_transcript_rows(df)

    if 'exon_id' not in df:
        logging.info("Creating 'exon_id' column")
        df = reconstruct_exon_id_column(df)

    return df



def parsearguments():
    epilog = """============================== Alea jacta est =============================="""
    parser = argparse.ArgumentParser(epilog=epilog,formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('INFNAME',help="Path to the xam or fastx file containing the reads")
    parser.add_argument('OUTFNAME',help="Output file name (fasta only).")

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    return(vars(parser.parse_args()))

if __name__ == "__main__":
    import argparse,sys,collections,os
    from tstk.io import openfastx,parsefastx
    from tstk.common import revcomp
    import pysam

    args = parsearguments()

    fpathnoext, fext = os.path.splitext(args["INFNAME"])

    if fext in [".bam",".sam"]:
        entries = [entry for entry in pysam.AlignmentFile(args["INFNAME"],"rb") if dict(entry.get_tags())["HI"] == 1]# only get the first entry in case of multi-mappers
        entries = [revcomp(str(e.seq)) if e.is_reverse else str(e.seq) for e in entries]
        counter = collections.Counter(entries) #only include one hit from the multi mappers
    else:
        fpathnoext,fext,ftype,fh = openfastx(args["INFNAME"])
        seqs = [str(seq) for name,seq,qual in parsefastx(fh)]
        counter = collections.Counter(seqs)
        fh.close()

    ofpathnoext,ofext,oftype,ofh = openfastx(args["OUTFNAME"].replace("fastq","fasta"),mode='wt')

    seqid = 1
    for seq in counter.most_common():
        ofh.write(">{}-{}\n{}\n".format(seqid,seq[1],seq[0]))
        seqid += 1

    ofh.close()
